/*
 * Copyright 2005,2006 Red Hat, Inc.
 *
 * Author: Chris Feist <cfeist@redhat.com>
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <unistd.h>
#include <netdb.h>
#include <errno.h>
#include <ldap.h>
#include <rpc/rpc.h>
#include <rpc/types.h>
#include <rpcsvc/yp.h>
#include <rpcsvc/ypclnt.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "automount.h"
#include "rpc_subs.h"

#define MODPREFIX "nsswitch: "

#define MAPTYPE_FILE 1
#define MAPTYPE_PROGRAM 2

int underscore_to_dot = 0;
static const struct timeval RPCTIMEOUT = {5, 0};

static int isfilemap(const char *);
static int isypmap(const char *);
static int isldapmap(const char *);

char *make_mapname(const char *loc, const char *type)
{
	int retsize, needs_etc = 0;
	char *retval;

	retsize = strlen(loc) + strlen(type) + 2;
       	if (*loc != '/') {
		retsize += 5; /* " "/etc/" */
		needs_etc = 1;
	}

	retval = malloc(retsize);
	if (!retval)
		return NULL;

	snprintf(retval, retsize, "%s:%s%s", type,
		 needs_etc ? "/etc/" : "", loc);

	return retval;
}

/*
 * Function which takes in a partial map name (ie. auto.misc), parses the
 * nsswitch.conf file and returns a valid map name (ie. yp:auto.misc or
 * file:/etc/auto.misc).  "cur" is the current map we are reading.  This
 * is used when determining the source of an included map.  For example.
 * if we are parsing /etc/auto.master, and it has the line "+auto.master",
 * we need to grab the auto.master from nis, if nis is specified in
 * nsswitch.conf.
 */
char *get_nsswitch_map(const char *loc, char *cur)
{
	char buf[1024];
	char *ordering;
	const char *automount_str = "automount:";
	char *comment = NULL;
	FILE *nsswitch;
	int found_automount = 0;
	char *retval = NULL;
	int retsize = 0;
	char *ypmapname = (char *)loc;

	debug(MODPREFIX "called nsswitch with: '%s' '%s'", loc, cur);
	nsswitch = fopen(_PATH_NSSWITCH_CONF, "r");
	if (!nsswitch) {
		error(MODPREFIX "Unable to open %s", _PATH_NSSWITCH_CONF);
		return NULL;
	}

	while ((ordering = fgets((char *)buf, sizeof(buf), nsswitch))) {
		if ((comment = strchr(ordering,'#')))
			*comment = '\0';
		while (isspace(*ordering)) ordering++;
		if (!strncmp(ordering, automount_str, sizeof(automount_str))) {
			ordering += strlen(automount_str);
			found_automount = 1;
			break;
		}
	}

	fclose(nsswitch);

	if (!found_automount)
		ordering = "files";

	/*
	 *  We support an option to convert underscores to dots when looking
	 *  up yp maps.
	 */
	if (underscore_to_dot) {
		char *u;

		ypmapname = strdup(loc);
		if (!ypmapname)
			return NULL;
		while ((u = strchr(ypmapname, '_')))
			*u = '.';
	}

	while (*ordering != '\0') {
		while (isspace(*ordering) && (*ordering != '\0')) ordering++;
		if (!strncmp(ordering, "files", 5)) {
			switch (isfilemap(loc)) {
			case MAPTYPE_FILE:
				retval = make_mapname(loc, "file");
				debug("comparing %s to %s\n", retval+5,cur);
				if (cur && !strcmp(cur, retval+5)) {
					free(retval);
					retval = NULL;
					break;
				}
				return retval;
			case MAPTYPE_PROGRAM:
				retval = make_mapname(loc, "program");
				debug("comparing %s to %s\n", retval+5,cur);
				if (cur && !strcmp(cur, retval+5)) {
					free(retval);
					retval = NULL;
					break;
				}
				return retval;
			default: // filemap doesn't exist
				break;
			}

		} else if ((!strncmp(ordering, "yp", 2) ||
			    !strncmp(ordering,"nis", 3)) &&
			   isypmap(ypmapname)) {
			retsize = strlen(ypmapname) + 4;
			retval = malloc(retsize);
			snprintf(retval, retsize, "yp:%s", ypmapname);
			return retval;
		} else if (!strncmp(ordering, "ldap", 4) && isldapmap(loc)) {
			retsize = strlen(loc) + 6;
			retval = malloc(retsize);
			snprintf(retval, retsize, "ldap:%s", loc);
			return retval;
		}
		while (!isspace(*ordering) && (*ordering != '\0')) ordering++;
	}
	error(MODPREFIX "couldn't find map %s", loc);
	return retval;
}

/*
 * Function takes in a filename and tests if it exists in "/etc/"
 * Returns: MAPTYPE_FILE if it is not executable, MAPTYPE_PROGRAM if it
 * is executable and 0 if it doesn't exists or has incorrect permissions.
 */

static int isfilemap(const char *loc)
{
	struct stat st;
	int ret = 0;	
	char *realfilemap;

	if (*loc != '/') {
		realfilemap = malloc(strlen(loc) + 6); /* '/etc/' + '\0' */
		if (!realfilemap) {
			crit(MODPREFIX "malloc failed.");
			return 0;
		}
		snprintf(realfilemap, strlen(loc) + 6, "/etc/%s", loc);
		ret = stat(realfilemap, &st);
		free(realfilemap);
	} else
		ret = stat(loc, &st);

	if (!ret) {
		if (st.st_uid != 0) {
			error(MODPREFIX "file /etc/%s exists but is not"
					" owned by root.", loc);
			return 0;
		} else if (st.st_mode & S_IRUSR) {
			if (st.st_mode & S_IXUSR) 
				return MAPTYPE_PROGRAM;
			else
				return MAPTYPE_FILE;
		}
	}
	debug(MODPREFIX "stat(%s) returned errno %d\n", loc, errno);
	return 0;
}

/*
 * Stolen from libc's ypclnt.c::yp_bind_ypbindprog
 */
static int autofs_yp_get_server(const char *domain, struct sockaddr_in *servaddr)
{
	struct sockaddr_in clnt_saddr;
	struct ypbind_resp ypbr;
	int clnt_sock;
	CLIENT *client;

	crit("attempting to get the server address for yp domain %s", domain);
	clnt_saddr.sin_family = AF_INET;
	clnt_saddr.sin_port = 0;
	clnt_saddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	clnt_sock = RPC_ANYSOCK;
	client = clnttcp_create(&clnt_saddr, YPBINDPROG, YPBINDVERS,
				&clnt_sock, 0, 0);
	if (client == NULL)
		return YPERR_YPBIND;
	set_cloexec(clnt_sock);

	/* Check the port number -- should be < IPPORT_RESERVED.
	   If not, it's possible someone has registered a bogus
	   ypbind with the portmapper and is trying to trick us. */
	if (ntohs(clnt_saddr.sin_port) >= IPPORT_RESERVED) {
		clnt_destroy(client);
		return YPERR_YPBIND;
	}

	if (clnt_call(client, YPBINDPROC_DOMAIN,
		      (xdrproc_t) xdr_domainname, (caddr_t) &domain,
		      (xdrproc_t) xdr_ypbind_resp,
		      (caddr_t) &ypbr, RPCTIMEOUT) != RPC_SUCCESS) {
		clnt_destroy(client);
		return YPERR_YPBIND;
	}
	clnt_destroy(client);

	if (ypbr.ypbind_status != YPBIND_SUCC_VAL) {
		debug("YPBINDPROC_DOMAIN: %s\n",
		      ypbinderr_string(ypbr.ypbind_resp_u.ypbind_error));
		return YPERR_DOMAIN;
	}
	memset(servaddr, '\0', sizeof(*servaddr));
	servaddr->sin_family = AF_INET;
	memcpy(&servaddr->sin_port,
	       ypbr.ypbind_resp_u.ypbind_bindinfo.ypbind_binding_port,
	       sizeof(servaddr->sin_port));
	memcpy(&servaddr->sin_addr.s_addr,
	       ypbr.ypbind_resp_u.ypbind_bindinfo.ypbind_binding_addr,
	       sizeof(servaddr->sin_addr.s_addr));

	crit("Successfully got address %x", servaddr->sin_addr.s_addr);
	return 0;
}

/*
 * The following code is ripped from glibc, with the slight advantage
 * that it actually returns the results to the caller.
 */
static int autofs_yp_order(char *domainname,
			   const char *mapname, unsigned int *order)
{
	enum clnt_stat result;
	ypreq_nokey req;
	ypresp_order resp;
	static struct sockaddr_in servaddr;
	int cl_sock = RPC_ANYSOCK;
	CLIENT *client;
	int retry = 0;

	/*
	 * Cache the server address so we don't overload the ypbind daemon.
	 */
retry:
	if (servaddr.sin_addr.s_addr == 0 &&
	    autofs_yp_get_server(domainname, &servaddr) != 0)
		return -1;

	client = clntudp_create(&servaddr, YPPROG, YPVERS, RPCTIMEOUT,&cl_sock);
	if (!client) {
		/*
		 * If the cached server address fails, re-do the
		 * server lookup once.
		 */
		if (retry++)
			return -1;
		memset(&servaddr, 0, sizeof(servaddr));
		goto retry;
	}
	set_cloexec(cl_sock);

	req.domain = domainname;
	req.map = (char *)mapname;
	result = clnt_call(client, YPPROC_ORDER,
			   (xdrproc_t) xdr_ypreq_nokey, (caddr_t) &req,
			   (xdrproc_t) xdr_ypresp_order, (caddr_t) &resp,
			   RPCTIMEOUT);
	clnt_destroy(client);
	if (result == YPERR_SUCCESS)
		/* We cast to ypresp_val although the pointer could
		   also be of type ypresp_key_val or ypresp_master or
		   ypresp_order or ypresp_maplist.  But the stat
		   element is in a common prefix so this does not
		   matter.
		*/
		result = ypprot_err(resp.stat);

	if (result != YPERR_SUCCESS)
		return -1;

	*order = resp.ordernum;
	xdr_free((xdrproc_t) xdr_ypresp_order, (char *)&resp);
	return 0;
}

int get_yp_order_for_domain(char *domainname,
			    const char *mapname, unsigned int *order)
{
	int err;
	static int libc_broken = 0;

	if (libc_broken)
		goto fallback;

	err = yp_order(domainname, mapname, order);
	if (err != YPERR_SUCCESS)
		return -1;
	if (*order != 0)
		return 0;
	/*
	 * libc yp_order was broken for many releases.  So, implement
	 * our own if we get a map order of 0.
	 */
	libc_broken = 1;
fallback:
	return autofs_yp_order(domainname, mapname, order);
}

int get_yp_order(const char *mapname, unsigned int *order)
{
	int err;
	char *domainname;

	if ((err = yp_get_default_domain(&domainname)) != YPERR_SUCCESS) {
		error(MODPREFIX "unable to get default yp domain");
		return -1;
	}

	err = get_yp_order_for_domain(domainname, mapname, order);
	return err;
}

/*
 * Function takes in a yp map name and returns
 * 1 if it exists or 0 if it doesn't.
 *
 * Some of this code borrowed from ypcat
 */
static int isypmap(const char *loc)
{
	int err;
	char *domainname;
	char *hostname;
	unsigned int order;

	if ((err = yp_get_default_domain(&domainname)) != YPERR_SUCCESS) {
		error(MODPREFIX "unable to get default yp domain");
		return 0;
	}

	if ((err = get_yp_order_for_domain(domainname, loc, &order)) < 0) {
		/*
		 * Some nisplus servers, when running in nis compatibility
		 * mode, will not support yp_order.  So, we check for the
		 * existence of the map with the yp_master call.
		 */
		err = yp_master(domainname, loc, &hostname);
		if (err != YPERR_SUCCESS) {
			debug(MODPREFIX "unable to find map, %s in domain, %s",
			      loc, domainname);
			return 0;
		}
		free(hostname);
	}

	return 1;
}

/*
 *  boolean function: it returns 1 if 'mapname' exists in the ldap sever, and
 *  0 if it can't be found.
 */
static int ldap_map_exists(const char *mapname,
			   const char *class, const char *attr)
{
	LDAP *ld = NULL;
	LDAPControl *server = NULL, *client = NULL;
	LDAPMessage *messages = NULL, *entry = NULL;
	char *attrs[2];
	char filter[LINE_MAX] = "";
	int result, ret = 0;
	int c;

	/* Initialize the LDAP library, pointing it at the default server. */
	ld = ldap_init(NULL, LDAP_PORT);
	if (ld == NULL) {
		crit(MODPREFIX "error initializing LDAP\n");
		return 0;
	}

	/* Try to switch to protocol 3. */
	c = 3;
	if (ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &c)
	    != LDAP_SUCCESS) {
		/* Just retry with the default protocol version. */
		ldap_unbind(ld);
		ld = ldap_init(NULL, LDAP_PORT);
		if (ld == NULL) {
			crit(MODPREFIX "error initializing LDAP\n");
			return 0;
		}
	}

	/* Connect to the server anonymously. */
	result = ldap_simple_bind_s(ld, NULL, NULL);
	if (result != LDAP_SUCCESS) {
		crit(MODPREFIX "error binding to ldap server: %s\n",
		     ldap_err2string(result));
		goto out;
	}

	/* We only want the key and value attributes from entries which
	 * match the query we'll perform. */
	attrs[0] = LDAP_NO_ATTRS;
	attrs[1] = NULL;

	/* Set the filter to only find the map we're looking at. */
	snprintf(filter, sizeof(filter), "(&(objectclass=%s)(%s=%s))",
		 class, attr, mapname);

	debug(MODPREFIX "searching for map %s with filter %s\n",
	      mapname, filter);
	/* Perform a synchronous query to find the DN of the map. */
	result = ldap_search_ext_s(ld, NULL,
				   LDAP_SCOPE_SUBTREE,
				   filter,
				   attrs, FALSE,
				   &server, &client,
				   NULL,
				   LDAP_NO_LIMIT,
				   &messages);
	if (result != LDAP_SUCCESS)
		goto out;

	/* We expected only one match.  Pull it from the results list. */
	entry = ldap_first_entry(ld, messages);
	if (entry != NULL)
		ret = 1; /* success */

out:
	ldap_unbind(ld);
	if (server)
		ldap_control_free(server);
	if (client)
		ldap_control_free(client);
	if (messages)
		ldap_msgfree(messages);

	return ret;

}

static int isldapmap(const char *loc)
{
	/*
	 * search for the three schemas:
	 *   (&(objectclass=nisMap)(nisMapName=%s))
	 *   (&(objectclass=automountMap)(ou=%s))
	 *   (&(objectclass=automountMap)(automountMapName=%s))
	 */
	if (ldap_map_exists(loc, "nisMap", "nisMapName")) {
		debug(MODPREFIX "map %s found on the ldap server\n", loc);
		return 1;
	}
	if (ldap_map_exists(loc, "automountMap", "ou")) {
		debug(MODPREFIX "map %s found on the ldap server\n", loc);
		return 1;
	}
	if (ldap_map_exists(loc, "automountMap", "automountMapName")) {
		debug(MODPREFIX "map %s found on the ldap server\n", loc);
		return 1;
	}

	debug(MODPREFIX "map %s not found on the ldap server\n", loc);
	return 0;
}

#ifdef STANDALONE
int do_debug = 0;

void print_usage(const char *prog)
{
	fprintf(stderr, "Usage %s [-u] <mapname>\n", basename(prog));
}

int main(int argc, char **argv)
{
	char *map;
	int  opt;

	if (argc < 2) {
		print_usage(argv[0]);
		exit(1);
	}

	while ((opt = getopt(argc, argv, "u")) != -1) {
		switch (opt) {
		case 'u':
			underscore_to_dot = 1;
			break;
		case '?':
		default:
			print_usage(argv[0]);
			exit(1);

		}
	}

	map = get_nsswitch_map(argv[optind], (char *)"auto.master");
	if (!map)
		exit(2);

	fprintf(stdout, "%s\n", map);
	free(map);
	exit(0);
}
#endif
