#ident "$Id: lookup_yp.c,v 1.4 2004/05/10 12:44:30 raven Exp $"
/* ----------------------------------------------------------------------- *
 *   
 *  lookup_yp.c - module for Linux automountd to access a YP (NIS)
 *                automount map
 *
 *   Copyright 1997 Transmeta Corporation - All Rights Reserved
 *   Copyright 2001-2003 Ian Kent <raven@themaw.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge MA 02139,
 *   USA; either version 2 of the License, or (at your option) any later
 *   version; incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */

#include <stdio.h>
#include <malloc.h>
#include <errno.h>
#include <unistd.h>
#include <syslog.h>
#include <time.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rpc/rpc.h>
#include <rpc/xdr.h>
#include <rpcsvc/yp_prot.h>
#include <rpcsvc/ypclnt.h>

#define MODULE_LOOKUP
#include "automount.h"

#define MAPFMT_DEFAULT "sun"

#define MODPREFIX "lookup(yp): "

struct lookup_context {
	const char *domainname;
	const char *mapname;
	struct parse_mod *parse;
};

struct callback_data {
	int status;
	int map;
	const char *root;
	char direct_base[KEY_MAX_LEN + 1];
	const char *name;
	int name_len;
	unsigned long type;
	const char *mapname;
	time_t age;
	struct lookup_context *context;
};

int lookup_version = AUTOFS_LOOKUP_VERSION;	/* Required by protocol */

/*
 *  Create a lookup context for the given map.
 */
struct lookup_context *context_init(const char *map)
{
	struct lookup_context *ctxt;
	int err;

	if (map == NULL) {
		crit(MODPREFIX "No map name");
		return NULL;
	}

	if (!(ctxt = malloc(sizeof(struct lookup_context)))) {
		crit(MODPREFIX "%m");
		return NULL;
	}
	memset(ctxt, 0, sizeof(*ctxt));

	ctxt->mapname = map;

	/* This should, but doesn't, take a const char ** */
	err = yp_get_default_domain((char **) &ctxt->domainname);
	if (err) {
		crit(MODPREFIX "map %s: %s\n", ctxt->mapname,
				yperr_string(err));
		free(ctxt);
		return NULL;
	}

	return ctxt;
}

int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **context)
{
	struct lookup_context *ctxt;

	if (argc < 1) {
		crit(MODPREFIX "No map name");
		return 1;
	}

	ctxt = context_init(argv[0]);
	if (!ctxt) {
		crit(MODPREFIX "%m");
		return 1;
	}
	*context = ctxt;

	if (!mapfmt)
		mapfmt = MAPFMT_DEFAULT;

	cache_init();

	return !(ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1));
}

int yp_all_callback(int status, char *ypkey, int ypkeylen,
		    char *val, int vallen, char *ypcb_data)
{
	struct callback_data *cbdata = (struct callback_data *) ypcb_data;
	const char *root = cbdata->root;
	time_t age = cbdata->age;
	char *key;
	char *mapent;

	if (status != YP_TRUE)
		return status;

	key = alloca(ypkeylen + 1);
	strncpy(key, ypkey, ypkeylen);
	*(key + ypkeylen) = '\0';

	mapent = alloca(vallen + 1);
	strncpy(mapent, val, vallen);
	*(mapent + vallen) = '\0';

	cache_update(root, key, mapent, age);

	return 0;
}

static int read_map(const char *root, time_t age, struct lookup_context *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	struct ypall_callback ypcb;
	struct callback_data ypcb_data;
	int err;

	ypcb_data.root = root;
	ypcb_data.age = age;

	ypcb.foreach = yp_all_callback;
	ypcb.data = (char *) &ypcb_data;

	err = yp_all((char *) ctxt->domainname, (char *) ctxt->mapname, &ypcb);

	if (err != YPERR_SUCCESS) {
		warn(MODPREFIX "read_map for %s failed: %s",
		       root, yperr_string(err));
		return 0;
	}

	/* Clean stale entries from the cache */
	cache_clean(root, age);

	return 1;
}

int lookup_ghost(const char *root, int ghost, time_t now, void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	time_t age = now ? now : time(NULL);
	struct mapent_cache *me;
	int status = 1;

	if (!read_map(root, age, ctxt))
		return LKP_FAIL;

	status = cache_ghost(root, ghost, ctxt->mapname, "yp", ctxt->parse);

	me = cache_lookup_first();
	/* me NULL => empty map */
	if (me == NULL)
		return LKP_EMPTY;

	if (*me->key == '/' && *(root + 1) != '-') {
		me = cache_partial_match(root);
		/* me NULL => no entries for this direct mount root or indirect map */
		if (me == NULL)
			return LKP_EMPTY | LKP_INDIRECT;
	}

	return status;
}

/*
 *  Return Values:
 *    CHE_OK      -  Success
 *    CHE_FAIL    -  malloc or yp_get_default_domain failed
 *    CHE_UPDATED -  either the key was newly added to the cache, or the
 *                   value associated with key changed.
 *    CHE_MISSING -  the key does not exist in the nis map
 *    yp error code - negated error returned from yp_match.
 */
int lookup_one(const char *root, const char *key, int key_len,
	       void *lookup_context)
{
	struct lookup_context *ctxt = (struct lookup_context *)lookup_context;
	char *mapent;
	int mapent_len;
	time_t age = time(NULL);
	int err;

	/*
	 * For reasons unknown, the standard YP definitions doesn't
	 * define input strings as const char *.  However, my
	 * understanding is that they will not be modified by the
	 * library.
	 */
	err = yp_match((char *) ctxt->domainname, (char *) ctxt->mapname,
		       (char *) key, key_len, &mapent, &mapent_len);

	if (err != YPERR_SUCCESS) {
		if (err == YPERR_KEY)
			return CHE_MISSING;

		return -err;
	}

	return cache_update(root, key, mapent, age);
}

static int lookup_wild(const char *root, struct lookup_context *ctxt)
{
	char *mapent;
	int mapent_len;
	time_t age = time(NULL);
	int err;

	mapent = alloca(MAPENT_MAX_LEN + 1);
	if (!mapent)
		return 0;

	err = yp_match((char *) ctxt->domainname,
		       (char *) ctxt->mapname, "*", 1, &mapent, &mapent_len);

	if (err == YPERR_SUCCESS)
		return cache_update(root, "*", mapent, age);
	
	return -err;
}

int lookup_mount(const char *root, const char *name, int name_len, void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	char key[KEY_MAX_LEN + 1];
	int key_len;
	char *mapent;
	int mapent_len;
	struct mapent_cache *me, *exists;
	int ret;
	time_t now = time(NULL);
	time_t t_last_read;
	int rmdir_if_mount_fails = 0;

	debug(MODPREFIX "looking up %s", name);

	key_len = format_key(key, KEY_MAX_LEN + 1, ap.type, root, name);
	if (key_len > KEY_MAX_LEN)
		return 1;

	/*
	 *  check map and if change is detected re-read map
	 */

	/* First check to see if this entry exists in the cache */
	exists = cache_lookup_distinct(key);
	ret = lookup_one(root, key, key_len, ctxt);
	if (ret == CHE_FAIL)
		return 1;

	/* was there a yp error? If so, simply use the cached copy. */
	if (ret < 0)
		warn(MODPREFIX 
		     "lookup for %s failed: %s", name, yperr_string(-ret));

	me = cache_lookup_first();
	t_last_read = me ? now - me->age : ap.exp_runfreq + 1;

	if (ret == CHE_UPDATED) {
		/* Have parent update its map */
		if (t_last_read > ap.exp_runfreq)
			kill(getppid(), SIGHUP);
	} else if (ret == CHE_MISSING && exists) {
		if (!cache_delete(root, key, 0)) {
			rmdir_if_mount_fails = 1;
		}

		/* Maybe update wild card map entry */
		if (ap.type == LKP_INDIRECT)
			lookup_wild(root, ctxt);

		/* Have parent update its map */
		if (t_last_read > ap.exp_runfreq)
			kill(getppid(), SIGHUP);
	}


	me = cache_lookup(key);
	if (me) {
		mapent = alloca(strlen(me->mapent) + 1);
		mapent_len = sprintf(mapent, "%s", me->mapent);
	} else {
		/* path component, do submount */
		me = cache_partial_match(key);
		if (me) {
			mapent = alloca(strlen(ctxt->mapname) + 20);
			mapent_len =
			    sprintf(mapent, "-fstype=autofs yp:%s", ctxt->mapname);
		}
	}

	if (!me) {
		error(MODPREFIX "key \"%s\" not found in map.", name);
		return EXIT_CODE_LOOKUP_FAILED;
	}

	mapent[mapent_len] = '\0';

	debug(MODPREFIX "%s -> %s", key, mapent);

	ret = ctxt->parse->parse_mount(root, name, name_len, mapent, ctxt->parse->context);

	if (ret && rmdir_if_mount_fails)
		rmdir_path(key);
	
	/* EXIT_CODE_MOUNT_FAILED is caught by the upper layers
	 * and used to populate the negative lookup cache */
	if (ret)
		ret = EXIT_CODE_MOUNT_FAILED;

	return ret;
}

int lookup_done(void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	int rv = close_parse(ctxt->parse);
	free(ctxt);
	cache_release();
	return rv;
}

/*
 *  Check for an entry in an included yp map.  Named lookup_one_included,
 *  as it is the included map counterpart to lookup_one.
 *
 *  Return Values:
 *    CHE_OK      -  Success
 *    CHE_FAIL    -  malloc or yp_get_default_domain failed
 *    CHE_MISSING -  key does not exist in the map
 *    CHE_UPDATED -  either the key was newly added to the cache, or the
 *                   value associated with key changed.
 */
int lookup_one_included(char *map, char *root, char *key, int key_len)
{
	struct lookup_context *ctxt;
	int rv = 0;

	ctxt = context_init(map);
	if (!ctxt)
		return CHE_FAIL;

	rv = lookup_one(root, key, key_len, ctxt);
	free(ctxt);
	return rv;
}

/*
 *  Check for the existence of "map", and read in its contents if it
 *  exists.  This is named lookup_map_included as it is the included map
 *  counterpart to read_map.
 *
 *  Return Values:
 *    1 - Success
 *    0 - Failure: malloc, yp_get_default_domain, or the yp lookup failed.
 */
int lookup_readmap_included(char *map, char *root, time_t age)
{
	struct lookup_context *ctxt;
	int rv = 0;

	ctxt = context_init(map);
	if (!ctxt)
		return 0;

	rv = read_map(root, age, ctxt);
	free(ctxt);
	return rv;
}
